import abc
from enum import Enum
from typing import Dict, Generic, List, Type, TypeVar, Union

import numpy as np
import pandas as pd
from datagen_protocol.schema.humans.human import Eyebrows, FacialHair

try:
    from collections import Sequence
except ImportError:
    from collections.abc import Sequence  # >=Python 3.10

from rich import pretty as pretty_rich

from datagen.api.assets import Background, DataSequence, Eyes, Glasses, Hair, Human, Mask
from datagen.api.catalog.attributes import AllOf, AnyOf, Exactly

Asset = TypeVar("Asset")


class AssetProvisioningHook(abc.ABC, Generic[Asset]):
    @abc.abstractmethod
    def __call__(self, *args) -> None:
        ...


class InitParametersHook(AssetProvisioningHook, abc.ABC, Generic[Asset]):
    @abc.abstractmethod
    def __call__(self, asset_id: str) -> dict:
        ...


class PostLoadHook(AssetProvisioningHook, abc.ABC, Generic[Asset]):
    @abc.abstractmethod
    def __call__(self, asset: Asset) -> None:
        ...


class AssetCreationHooks(Generic[Asset]):
    def __init__(self, hooks: List[AssetProvisioningHook]):
        self._hooks = hooks

    @property
    def _init_hooks(self) -> List[InitParametersHook]:
        return list(filter(lambda h: isinstance(h, InitParametersHook), self._hooks))  # type: ignore

    def get_init_params(self, asset_id: str) -> dict:
        init_params = {}
        for h in self._init_hooks:
            init_params.update(**h(asset_id))
        return init_params

    @property
    def _post_load_hooks(self) -> List[PostLoadHook]:
        return list(filter(lambda h: isinstance(h, PostLoadHook), self._hooks))  # type: ignore

    def post_load(self, asset: Asset) -> None:
        for h in self._post_load_hooks:
            h(asset)


class AssetInstancesProvisioner(Generic[Asset]):
    def __init__(
        self,
        asset_type: Type[Asset],
        asset_id_to_asset_attrs: dict,
        hooks: List[AssetProvisioningHook[Asset]],
    ):
        self._asset_type = asset_type
        self._asset_id_to_asset_attrs = asset_id_to_asset_attrs
        self._hooks = AssetCreationHooks(hooks)

    def provision(self, asset_id: str) -> Asset:
        asset = self._asset_type(
            id=asset_id,
            **self._hooks.get_init_params(asset_id),
            attributes=self._asset_id_to_asset_attrs[asset_id],
        )
        return asset

    def parse(self, asset_id: str, asset_body_dict: dict) -> Asset:
        asset = self._asset_type(
            id=asset_id,
            **asset_body_dict,
            attributes=self._asset_id_to_asset_attrs[asset_id],
        )
        return asset

    def post_load(self, asset: Asset) -> Asset:
        self._hooks.post_load(asset)
        return asset


class CatalogInstancesCache(Generic[Asset]):
    def __init__(self, provisioner: AssetInstancesProvisioner):
        self._provisioner = provisioner
        self._cache = {}

    def get(self, asset_id: str) -> Asset:
        asset = self._load(asset_id)
        asset = self._provisioner.post_load(asset)
        return asset

    def _load(self, asset_id) -> Asset:
        try:
            asset = self._cache[asset_id].copy(deep=True)
        except KeyError:
            asset = self._cache.setdefault(asset_id, self._provisioner.provision(asset_id))
        return asset


class AssetInstancesList(Sequence):
    def __init__(self, instances_cache: CatalogInstancesCache, assets_ids: List[str]):
        self._instances_cache = instances_cache
        self._assets_ids = assets_ids

    def __repr__(self):
        return f"<{self.__class__.__name__}({len(self)})>"

    def __iter__(self):
        return AssetInstancesIter(self)

    def __len__(self):
        return len(self._assets_ids)

    def __getitem__(self, index):
        return self._instances_cache.get(self._assets_ids[index])

    def __contains__(self, element):
        return element in self._assets_ids


class AssetInstancesIter:
    def __init__(self, instances_list: AssetInstancesList):
        self._idx = 0
        self._instances_list = instances_list

    def __iter__(self):
        return self

    def __next__(self):
        if self._idx < len(self._instances_list):
            self._idx += 1
            return self._instances_list[self._idx - 1]
        else:
            raise StopIteration


class AttributesDataFrame:
    def __init__(self, attributes_df: pd.DataFrame):
        self._df = attributes_df

    def query(self, attributes: Dict[str, Union[Enum, AllOf, AnyOf, Exactly]], limit: int = None) -> List[str]:
        if attributes:
            df = self._df.loc[self._create_query_series(attributes)]
        else:
            df = self._df
        return list(df.index)[:limit]

    def _create_query_series(self, attributes: Dict[str, Union[Enum, AllOf, AnyOf, Exactly]]) -> pd.Series:
        attrs_series_list = [
            self._create_attr_series(attr_name, attr_query_val) for attr_name, attr_query_val in attributes.items()
        ]
        return np.logical_and.reduce(attrs_series_list)  # type: ignore

    def _create_attr_series(self, attr_name: str, attr_query_val: Union[str, Enum, AllOf, AnyOf, Exactly]) -> pd.Series:
        if isinstance(attr_query_val, (str, Enum)):
            return self._create_series(attr_name, attr_query_val)
        elif isinstance(attr_query_val, (AllOf, AnyOf)):
            series_list = [self._create_series(attr_name, attr_val) for attr_val in attr_query_val]
            operator = np.logical_and if isinstance(attr_query_val, AllOf) else np.logical_or
            return operator.reduce(series_list)  # type: ignore

    def _create_series(self, attr_name: str, attr_val: Union[str, Enum]) -> pd.Series:
        if isinstance(attr_val, Enum):
            attr_val = attr_val.value
        return self._df[f"{attr_val}_{attr_name}"].__eq__(True)

    @staticmethod
    def from_dict(asset_id_to_asset_attrs: dict) -> "AttributesDataFrame":
        attrs_dataframe_dict = {}
        for asset_id, asset_attrs_dict in asset_id_to_asset_attrs.items():
            attrs_dataframe_dict[asset_id] = AttributesDataFrame._get_dataframe_row(asset_attrs_dict)
        attr_dataframe = pd.DataFrame.from_dict(attrs_dataframe_dict, orient="index")
        return AttributesDataFrame(attr_dataframe)

    @staticmethod
    def _get_dataframe_row(asset_attrs_dict: dict) -> dict:
        dataframe_row = {}
        for attr_name, attr_value in asset_attrs_dict.items():
            if not isinstance(attr_value, list):
                attr_value = [attr_value]
            for v in attr_value:
                if isinstance(v, str):
                    dataframe_row[v + "_" + attr_name] = True
        return dataframe_row


class AssetCatalog(Generic[Asset]):
    def __init__(
        self,
        asset_type: Type[Asset],
        asset_id_to_asset_attrs: dict,
        hooks: List[AssetProvisioningHook[Asset]] = [],
    ):
        self._provisioner = AssetInstancesProvisioner(asset_type, asset_id_to_asset_attrs, hooks)
        self._instances_cache = CatalogInstancesCache(self._provisioner)
        self._attributes_df = AttributesDataFrame.from_dict(asset_id_to_asset_attrs)

    def get(self, id: str = None, limit: int = None, **attributes) -> Union[Asset, List[Asset]]:
        if id:
            return self._instances_cache.get(id)
        else:
            matching_assets_ids = self._attributes_df.query(attributes, limit)
            if limit == 1:
                return self._instances_cache.get(matching_assets_ids[0])
            else:
                return AssetInstancesList(self._instances_cache, matching_assets_ids)

    def count(self, **attributes) -> int:
        return len(self._attributes_df.query(attributes))

    def parse(self, id: str, **asset_body) -> Asset:
        return self._provisioner.parse(asset_id=id, asset_body_dict=asset_body)


class BackgroundsCatalog(AssetCatalog[Background]):
    def get(
        self,
        id: str = None,
        limit: int = None,
        # generator=Generator.IDENTITIES,
        **attrs,
    ) -> Union[Asset, List[Asset]]:
        return super().get(
            id,
            limit,
            # generator=generator,
            **attrs,
        )


class DatagenAssetsCatalog:
    def __init__(
        self,
        humans: AssetCatalog[Human],
        sequences: AssetCatalog[DataSequence],
        hair: AssetCatalog[Hair],
        eyes: AssetCatalog[Eyes],
        eyebrows: AssetCatalog[Eyebrows],
        facial_hair: AssetCatalog[FacialHair],
        glasses: AssetCatalog[Glasses],
        masks: AssetCatalog[Mask],
        backgrounds: BackgroundsCatalog,
    ):
        self._humans = humans
        self._sequences = sequences
        self._hair = hair
        self._eyes = eyes
        self._eyebrows = eyebrows
        self._facial_hair = facial_hair
        self._glasses = glasses
        self._masks = masks
        self._backgrounds = backgrounds

        pretty_rich.install()

    @property
    def humans(self) -> AssetCatalog[Human]:
        return self._humans

    @property
    def sequences(self) -> AssetCatalog[DataSequence]:
        return self._sequences

    @property
    def hair(self) -> AssetCatalog[Hair]:
        return self._hair

    @property
    def eyes(self) -> AssetCatalog[Eyes]:
        return self._eyes

    @property
    def eyebrows(self) -> AssetCatalog[Eyebrows]:
        return self._eyebrows

    @property
    def facial_hair(self) -> AssetCatalog[FacialHair]:
        return self._facial_hair

    @property
    def beards(self) -> AssetCatalog[FacialHair]:
        return self._facial_hair

    @property
    def glasses(self) -> AssetCatalog[Glasses]:
        return self._glasses

    @property
    def masks(self) -> AssetCatalog[Mask]:
        return self._masks

    @property
    def backgrounds(self) -> BackgroundsCatalog:
        return self._backgrounds



import marshmallow_dataclass

from datagen.modalities.textual.common.ndarray import NumpyArray


@marshmallow_dataclass.dataclass
class SixDOF:
    location: NumpyArray
    look_at_vector: NumpyArray



import asyncio
import os
import tarfile
from http import HTTPStatus
from pathlib import Path
from typing import Any, List, Tuple

import aiofiles
from aiohttp import ClientPayloadError

from datagen.api.client.schemas import DataResponse, DataResponseStatus, DownloadURL
from datagen.api.client.session import ClientSession, SessionsResponse
from datagen.core.tasks.task import ClientTask, Task
from datagen.dev.logging import get_logger

logger = get_logger(__name__)


class StopGenerationTask(ClientTask):
    """
    A task that stops a data generation process.

    This task sends an HTTP POST data to the Datagen services to stop the data generation
    process with the specified generation id.

    In order to use this task you need to have previously started a data generation process, so you need to pass a
    generation id as input.
    Input for this task is a generation_id.
    """

    def __init__(
        self,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)

        self._url = "/v1/generations/{generation_id}/stop"

    async def execute(self) -> None:
        generation_id = await self._get_input()
        session = self._session.session
        async with session.post(url=self._url.format(generation_id=generation_id)) as resp:
            self._handle_response(
                response=SessionsResponse(status_code=resp.status, payload=await resp.json()),
                generation_id=generation_id,
            )
            await self.output_channel.put(True)

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return f"Failed to stop generation: {kwargs['generation_id']} with error: {error_msg}"

    def _success_message(self, **kwargs) -> str:
        return f"Successfully stopped data generation: {kwargs['generation_id']}."


class StatusTask(ClientTask):
    """
    A task that queries the current generation status.

    Generation status consists of: time estimation in ms, process percentage and the status of the process
    (e.g. STARTED, IN-PROGRESS, etc.)

    This task sends an HTTP GET data to the Datagen services to query the data generation
    status with the specified generation id.

    In order to use this task you need to have previously started a data generation process, so you need to pass a
    generation id as input.

    Input: generation_id.
    Output: GenerationStatus
    """

    def __init__(
        self,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self._url = "/v1/generations/{generation_id}/status"

    async def execute(self) -> None:
        generation_id = await self._get_input()
        session = self._session.session
        async with session.get(url=self._url.format(generation_id=generation_id)) as resp:
            response = SessionsResponse(status_code=resp.status, payload=await resp.json())
            self._handle_response(response=response, generation_id=generation_id)
            await self.output_channel.put(DataResponseStatus(**response.payload))

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return f"Failed to receive generation status of: {kwargs['generation_id']} with error: {error_msg}"

    def _success_message(self, **kwargs) -> str:
        return f"Successfully received generation status of: {kwargs['generation_id']}."


class GetDownloadURLsTask(ClientTask):
    """
    A task that gets the download URLs for the output files of a data generation process.

    The URLs will have an expiration date. In case of an expired date: please use this task again to gen new URLs.

    This task sends an HTTP GET data to Datagen services to get the download URLs for the output files of the data
    generation process with the specified generation id.

    If the generation is still in progress, an empty list will be returned.
    Input: generation_id.
    Output: list of urls to files.
    """

    def __init__(
        self,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self._url = "/v1/generations/{generation_id}/download"

    async def execute(self) -> None:
        generation_id = await self._get_input()
        session = self._session.session
        async with session.get(url=self._url.format(generation_id=generation_id)) as resp:
            response = SessionsResponse(status_code=resp.status, payload=await resp.json())
            if response.status_code == HTTPStatus.ACCEPTED:
                logger.info(
                    f"Data generation is in progress. Download URLs for generation ID "
                    f"{generation_id} will soon be ready."
                )
            else:
                self._handle_response(response=response, generation_id=generation_id)

            response = [DownloadURL.parse_url(url=url) for url in response.payload]
            await self.output_channel.put(response)

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return f"Failed to get download URLs of: {kwargs['generation_id']} with error: {error_msg}"

    def _success_message(self, **kwargs) -> str:
        return f"Successfully received download URLs of: {kwargs['generation_id']}."


class InitDataGenerationTask(ClientTask):
    """
    A task that initializes a data generation request process.
    For data with 2000 datapoints or less, use DataGenerationTask.

    This task sends an HTTP GET data to Datagen services to initialize a multi-stage data generation process.
    If request was successful, you will receive a generation id that you'll pass forward to UploadRequestTask and
    FinalizeDataGenerationTask.
    """

    def __init__(
        self,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self._url = "/v1/generations"

    async def execute(self) -> None:
        request, name = await self._get_input()
        session = self._session.session
        async with session.post(url=self._url, json={"title": name}) as resp:
            response = SessionsResponse(status_code=resp.status, payload=await resp.json())
            self._handle_response(response=response)
            generation_id = response.payload["generation_id"]
            await self.output_channel.put([(generation_id, batch) for batch in request])

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return f"Failed to initialize generation request with error: {error_msg}"

    def _success_message(self, **kwargs) -> str:
        return "Successfully initialized generation request."


class UploadRequestTask(ClientTask):
    """
    A task that uploads a data generation request.

    This task sends an HTTP POST data to Datagen services to initialize that a multi-stage data generation process.
    This tasks depend on InitDataGenerationTask's output.

    Input: generation_id, DataRequest (a batch with datapoints < 2000)
    Output: DataRequest object with dgu-hour cost, number of datapoints, number od scenes
    """

    def __init__(
        self,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self._url = "/v1/generations/{generation_id}"

    async def execute(self) -> None:
        generation_id, request = await self._get_input()
        session = self._session.session
        async with session.put(url=self._url.format(generation_id=generation_id), json=request.dict()) as resp:
            response = SessionsResponse(status_code=resp.status, payload=await resp.json())
            self._handle_response(response=response, generation_id=generation_id)
            await self.output_channel.put(DataResponse(**response.payload))

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return (
            f"Failed to upload generation request for generation id: {kwargs['generation_id']} with error: {error_msg}"
        )

    def _success_message(self, **kwargs) -> str:
        return f"Successfully uploaded generation request of generation id: {kwargs['generation_id']}."


class FinalizeDataGenerationTask(ClientTask):
    """
    This is the third and last part of the multipart request upload.
    Once you uploaded your requests with UploadRequestTask, use this task to trigger the data generation process.

    This task sends an HTTP POST data to Datagen services to start the data generation process itself.
    This tasks depend on UploadRequestTask's output.

    Input: generation_id
    Output: DataRequest object with dgu-hour cost, number of datapoints, number od scenes
    """

    def __init__(self, session: ClientSession, **kwargs):
        super().__init__(session=session, **kwargs)
        self._url = "/v1/generations/{generation_id}"

    async def execute(self) -> None:
        generation_id = await self._get_input()
        session = self._session.session
        async with session.post(url=self._url.format(generation_id=generation_id)) as resp:
            response = SessionsResponse(status_code=resp.status, payload=await resp.json())
            self._handle_response(response=response, generation_id=generation_id)
            await self.output_channel.put(DataResponse(**response.payload))

    async def _get_input(self) -> Any:
        upload_response = await self.input_channel.take()
        return upload_response[0].generation_id

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return (
            f"Failed to finalize generation request for generation id: {kwargs['generation_id']} "
            f"with error: {error_msg}."
        )

    def _success_message(self, **kwargs) -> str:
        return f"Successfully finalized generation request for generation id: {kwargs['generation_id']}."


class DownloadFileTask(ClientTask):
    """
    This dask downloads a single file to a requested dictionary.
    """

    def __init__(self, session: ClientSession, **kwargs):
        super().__init__(session=session, **kwargs)
        self._lock = asyncio.Lock()

    async def execute(self) -> None:
        download_url = await self._get_input()
        session = self._session.session
        logger.info(f"Starting download from {download_url.url}. The file will be located at {download_url.filename}")
        async with session.get(url=download_url.url) as resp:
            if resp.status != HTTPStatus.OK:
                logger.error(self._error_message(error_msg="Download failed.", url=download_url.url))
                resp.raise_for_status()
            async with aiofiles.open(download_url.filename, "wb") as file:
                chunk_size = 10 * 1024 * 1024  # 10MB
                try:
                    async for chunk in resp.content.iter_chunked(chunk_size):
                        await file.write(chunk)
                except ClientPayloadError as e:
                    print(e)

            logger.info(self._success_message(url=download_url.url, filename=download_url.filename))
            await self.output_channel.put((download_url.filename, download_url.dataset_name))

    def _error_message(self, error_msg: str, **kwargs) -> str:
        return f"Failed to download {kwargs['url']} with error: {error_msg}."

    def _success_message(self, **kwargs) -> str:
        return f"Successfully downloaded {kwargs['url']} to {kwargs['filename']}."


class ExtractFilesTask(Task):
    """
    This task receives a list of files, merges them into 1 file and extracts it to the requested directory.
    """

    def __init__(self, remove_tar_files: bool = True, **kwargs):
        super().__init__(**kwargs)
        self._remove_tar_files = remove_tar_files

    async def execute(self) -> None:
        paths, compressed_file, dest_file = await self._get_input()
        if len(paths) > 1:
            ExtractFilesTask._merge_file(paths=paths, merged_tar=compressed_file)

        ExtractFilesTask._extract(src=compressed_file, dst=dest_file)

        if self._remove_tar_files:
            ExtractFilesTask._remove_files(paths=paths)

        await self.output_channel.put(dest_file)

    async def _get_input(self) -> Any:
        file_paths = await self.input_channel.take()
        dataset_names = set(name for _, name in file_paths)
        if len(dataset_names) != 1:
            raise ValueError("Got invalid download links. all links must relate to the same dataset.")
        return ExtractFilesTask._export_urls(file_paths=file_paths)

    @staticmethod
    def _export_urls(file_paths: Tuple[Any, ...]) -> Tuple[Tuple[Any, ...], str, str]:
        urls = tuple(url for url, _ in file_paths)
        if len(file_paths) == 1:
            compressed_file = str(Path(file_paths[0][0]))
            dest_file = str(Path(file_paths[0][0]).parent.joinpath(file_paths[0][1]))
            return urls, compressed_file, dest_file
        else:
            compressed_file = str(Path(file_paths[0][0]).parent.joinpath(file_paths[0][1] + ".tar.gz"))
            return urls, compressed_file, compressed_file.replace(".tar.gz", "")

    @staticmethod
    def _merge_file(paths: List[str], merged_tar: str) -> None:
        try:
            with tarfile.open(merged_tar, mode="w:gz") as merged_file:
                names = []
                for filename in paths:
                    with tarfile.open(filename, mode="r:gz") as file:
                        members = [m for m in file.getmembers() if m.name not in names]
                        names += file.getnames()
                        for member in members:
                            if member.isdir():
                                merged_file.addfile(member)
                            else:
                                merged_file.addfile(member, file.extractfile(member))

                logger.info(f"Successfully merged {len(paths)} files to {merged_tar}.")
        except Exception as e:
            logger.error(f"Failed to merge tar.gz files to {merged_file}.")
            raise e

    @staticmethod
    def _extract(src: str, dst: str) -> None:
        try:
            with tarfile.open(src, "r:gz") as tar:
                tar.extractall(path=dst)
                logger.info(f"Successfully extracted file to {dst}.")
        except Exception as e:
            logger.error(f"Failed to extract {dst}.")
            raise e

    @staticmethod
    def _remove_files(paths: List[str]) -> None:
        for path in paths:
            try:
                os.remove(path=path)
                logger.info(f"Successfully removed file: {path}.")
            except Exception as e:
                logger.error(f"Failed to removed {path} with error: {str(e)}.")



from dependency_injector import containers, providers

from datagen.api.client.containers import SessionContainer
from datagen.core.tasks.task import SessionTask, Task, TaskChain, TaskGroup
from datagen.core.tasks.task_definitions import (
    DownloadFileTask,
    ExtractFilesTask,
    FinalizeDataGenerationTask,
    GetDownloadURLsTask,
    InitDataGenerationTask,
    StatusTask,
    StopGenerationTask,
    UploadRequestTask,
)
from datagen.dev.logging import get_logger

logger = get_logger(__name__)


def create_data_generation_pipeline(task_container: containers.DeclarativeContainer, number_of_batches: int) -> Task:
    session = SessionContainer.client_session()
    init_task = task_container.task_factory.initialize(session=session)
    upload_tasks = task_container.collection_factory.task_group(
        tasks=[task_container.task_factory.upload(session=session) for _ in range(number_of_batches)],
    )
    finalize_task = task_container.task_factory.finalize(session=session)
    multipart_task = task_container.collection_factory.task_chain(tasks=[init_task, upload_tasks, finalize_task])
    session_task = SessionTask(
        task=multipart_task,
        session=session,
    )
    return session_task


def create_download_pipeline(
    task_container: containers.DeclarativeContainer, number_of_files: int, remove_tar_files: bool
) -> Task:
    session = SessionContainer.client_session(
        base_url=None, headers={"Connection": "keep-alive"}, timeout=SessionContainer.long_timeout()
    )
    download_task = task_container.collection_factory.task_group(
        tasks=[task_container.task_factory.download_file(session=session) for _ in range(number_of_files)],
    )
    extract_task = task_container.task_factory.extract_files(remove_tar_files=remove_tar_files)
    download_pipeline = task_container.collection_factory.task_chain(tasks=[download_task, extract_task])
    session_task = SessionTask(task=download_pipeline, session=session)
    return session_task


def create_client_task(task_container: containers.DeclarativeContainer, task_name: str, **kwargs) -> Task:
    session = SessionContainer.client_session()
    task = task_container.task_factory.providers[task_name](session=session, **kwargs)
    return SessionTask(task=task, session=session)


class TaskContainer(containers.DeclarativeContainer):

    __self__ = providers.Self()

    task_factory = providers.FactoryAggregate(
        stop_generation=providers.Factory(StopGenerationTask),
        get_download_links=providers.Factory(GetDownloadURLsTask),
        status=providers.Factory(StatusTask),
        initialize=providers.Factory(InitDataGenerationTask),
        upload=providers.Factory(UploadRequestTask),
        finalize=providers.Factory(FinalizeDataGenerationTask),
        extract_files=providers.Factory(ExtractFilesTask),
        download_file=providers.Factory(DownloadFileTask),
    )

    client_task = providers.Callable(create_client_task, task_container=__self__)

    collection_factory = providers.FactoryAggregate(
        task_chain=providers.Factory(TaskChain),
        task_group=providers.Factory(TaskGroup),
    )

    pipeline_factory = providers.FactoryAggregate(
        data_generation=providers.Callable(create_data_generation_pipeline, task_container=__self__),
        download=providers.Callable(create_download_pipeline, task_container=__self__),
    )



from datagen.dev import mutually_exclusive
from datagen.dev.logging import get_logger
from datagen.dev.modules import FunctionalModule, get_resource_path, load_resource
from datagen.dev.plugins import PluginsFactory



from dataclasses import field
from typing import Optional

import marshmallow_dataclass
from marshmallow import fields, pre_load

from datagen.modalities.textual.common.ndarray import NumpyArray


@marshmallow_dataclass.dataclass
class ObjectCenterOfGeometry:
    semantic_name: Optional[str]
    coords_2d: NumpyArray
    coords_3d: NumpyArray

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        center_of_mass_props = in_data.pop("center_of_mass")
        semantic_name = in_data["semantic_name"]
        coords_3d = center_of_mass_props["global_3d"]
        coords_2d = center_of_mass_props["pixel_2d"]
        coords_2d.pop("depth")
        return {"semantic_name": semantic_name, "coords_3d": coords_3d, "coords_2d": coords_2d}


class ObjectCenterOfGeometryField(fields.Field):
    def _deserialize(self, value, *args, **kwargs):
        return ObjectCenterOfGeometry.Schema().load(value)


@marshmallow_dataclass.dataclass
class CenterOfGeometry:
    dict_: dict = field(metadata={
        "marshmallow_field": fields.Dict(keys=fields.Str, values=ObjectCenterOfGeometryField)
    })

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        return {"dict_": in_data}




import dataclasses
from collections import namedtuple
from typing import List, TypeVar

import marshmallow
import marshmallow_dataclass
from marshmallow import pre_load
from marshmallow.fields import Field

SINGLE_ENV_DEFAULT_IMG_NAME = "visible_spectrum.png"


class Environment:
    def __init__(self, **attrs):
        for attr_name, attr_val in attrs.items():
            setattr(self, attr_name, attr_val)

    def matches(self, **env_attributes) -> bool:
        is_match = True
        for attr_name, attr_value in env_attributes.items():
            is_match &= self.__dict__[attr_name] == attr_value
        return is_match

    @property
    def transparent_background(self) -> bool:
        return getattr(self, "background", None) == "transparent"


EnvironmentType = TypeVar("EnvironmentType")


class EnvironmentField(Field):
    def _deserialize(self, anv_attrs, *args, **kwargs):
        environment = Environment(**anv_attrs)
        return environment


class EnvironmentsSchema(marshmallow.Schema):
    TYPE_MAPPING = {EnvironmentType: EnvironmentField}

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        if "environments" not in in_data:
            in_data["image_name"] = SINGLE_ENV_DEFAULT_IMG_NAME
            in_data = {"environments": [in_data]}
        return in_data


@marshmallow_dataclass.dataclass(base_schema=EnvironmentsSchema)
class Environments:
    environments: List[EnvironmentType]

    def __iter__(self):
        for environment in self.environments:
            yield environment

    def __getitem__(self, image_name: str):
        return next(env for env in self if env.image_name == image_name)



import asyncio
from abc import ABC, abstractmethod
from http import HTTPStatus
from typing import Any, List

from asyncio_channel import create_channel
from asyncio_channel._channel import Channel

from datagen.api.client.exceptions import HttpStatusHandler
from datagen.api.client.schemas import ErrorResponse, parse_error
from datagen.api.client.session import ClientSession, SessionsResponse
from datagen.dev.logging import get_logger

logger = get_logger(__name__)


class Task(ABC):
    """
    Abstract base class for tasks.

    A task represents a unit of work that can be executed asynchronously.
    Tasks can take input from a channel and produce output on another channel.
    """

    def __init__(self, **kwargs):
        self.input_channel = None
        self.output_channel = None

    def setup(self) -> None:
        self.input_channel = create_channel()
        self.output_channel = create_channel()

    @abstractmethod
    async def execute(self) -> None:
        pass

    async def _get_input(self) -> Any:
        return await self.input_channel.take()


class SessionTask(Task):
    """
    This task is used when you want tasks to share the same aiohttp.ClintSession.
    It is advised to run ClientTasks or a CollectionTask that contains ClientTasks inside a SessionTask
    for improved performance.
    """

    def __init__(
        self,
        task: Task,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._task = task
        self._session = session

    def setup(self) -> None:
        self._task.setup()
        self.input_channel = self._task.input_channel
        self.output_channel = self._task.output_channel

    async def execute(self) -> None:
        async with self._session:
            await self._task.execute()


class ClientTask(Task):
    """
    Base class for tasks that require an ClientSession.
    Use derived class of ClientTask only inside a SessionTask, e.g. the task input of a SessionTask will be a ClientTask
    or a CollectionTask that contains SessionTask.
    """

    def __init__(
        self,
        session: ClientSession,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._session = session
        self._url = ""

    def _handle_response(self, response: SessionsResponse, **kwargs) -> Any:
        if response.status_code not in [HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.ACCEPTED]:
            error = ErrorResponse(**parse_error(response.payload))
            HttpStatusHandler.handle(
                status_code=response.status_code, message=self._error_message(error_msg=str(error), **kwargs)
            )

        logger.info(self._success_message(**kwargs))

    @abstractmethod
    def _error_message(self, error_msg: str, **kwargs) -> str:
        pass

    @abstractmethod
    def _success_message(self, **kwargs) -> str:
        pass


class TaskCollection(Task):
    def __init__(self, tasks: List[Task], **kwargs):
        super().__init__(**kwargs)
        self._tasks = tasks

    def setup(self) -> None:
        super().setup()
        for task in self._tasks:
            task.setup()


class TaskChain(TaskCollection):
    """
    A chain of tasks to be executed one sequentially.
    """

    def __init__(self, tasks: List[Task], **kwargs):
        super().__init__(tasks=tasks, **kwargs)

    def setup(self) -> None:
        super().setup()
        for idx, task in enumerate(self._tasks):
            if idx == 0:
                task.input_channel = self.input_channel
            else:
                task.input_channel = self._tasks[idx - 1].output_channel

            if idx == len(self._tasks) - 1:
                task.output_channel = self.output_channel
            else:
                task.output_channel = self._tasks[idx + 1].input_channel

    async def execute(self) -> None:
        for task in self._tasks:
            await task.execute()


class TaskGroup(TaskCollection):
    """
    A group of tasks to be executed one asynchronously without a specific order.
    """

    def __init__(self, tasks: List[Task], **kwargs):
        super().__init__(tasks=tasks, **kwargs)

    async def execute(self) -> None:
        await self._get_input()
        await asyncio.gather(*(task.execute() for task in self._tasks))
        await self.output_channel.put(await self._get_output())

    async def _get_input(self) -> Any:
        batches = await self.input_channel.take()
        for batch, task in zip(batches, self._tasks):
            await task.input_channel.put(batch)

    async def _get_output(self) -> Any:
        return await asyncio.gather(*(task.output_channel.take() for task in self._tasks))

    def _get_src_channel(self) -> Channel:
        return self.input_channel



import os
from typing import ClassVar

from datagen import modalities
from datagen.components.datapoint.entity import base


class DataPoint(base.DataPoint):
    PLUGINS_GROUP_NAME: ClassVar[str] = "datagen.plugins.datapoints.hic"

    @modalities.textual_modality
    def actor_metadata(self) -> modalities.TextualModality:
        return modalities.TextualModality(factory_name="actors", file_name="actor_metadata.json")

    @modalities.textual_modality
    def keypoints(self) -> modalities.TextualModality:
        return modalities.TextualModality(
            factory_name="keypoints", file_name=os.path.join("key_points", "all_key_points.json")
        )

    @modalities.textual_modality
    def _center_of_geometry(self) -> modalities.TextualModality:
        return modalities.TextualModality(factory_name="center_of_geometry", file_name="center_of_geometry.json")

    @property
    def center_of_geometry(self) -> dict:
        return self._center_of_geometry.dict_



from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from datagen.components.camera import Camera
from datagen.components.datapoint import DataPoint


@dataclass
class Scene:
    path: Path
    cameras: List[Camera] = field(init=False, repr=False)

    def __post_init__(self):
        self.cameras = self._init_cameras()

    def _init_cameras(self) -> List[Camera]:
        return [Camera(name=camera_name, scene_path=self.path) for camera_name in sorted(self._get_cameras_names())]

    def _get_cameras_names(self) -> List[str]:
        if Scene.is_path_to_hic_scene(self.path):
            # for HIC, we have to select a random frame to get the cameras names of from.
            cameras_base_path = self.path.joinpath("frames", "001")
        else:
            cameras_base_path = self.path
        cameras_dirs = filter(lambda path_: path_.is_dir(), cameras_base_path.iterdir())
        return [cam_dir.name for cam_dir in cameras_dirs]

    @staticmethod
    def is_path_to_hic_scene(scene_path: Path) -> bool:
        hic_frames_dir = scene_path.joinpath("frames")
        return hic_frames_dir.exists() and hic_frames_dir.is_dir()

    @property
    def datapoints(self) -> List[DataPoint]:
        return [datapoint for datapoint in self]

    def __getitem__(self, key):
        return self.datapoints[key]

    def __iter__(self):
        for camera in self.cameras:
            for datapoint in camera:
                yield datapoint

    def __len__(self):
        return len(self.datapoints)



from dependency_injector import containers, providers

from datagen.modalities.textual.base.containers import (
    BaseModalitiesContainer,
    modality_dataclass_factory,
    modality_pydantic_factory,
    modality_factory,
)


class ActorMetadataModalityContainer(containers.DeclarativeContainer):

    from .actor_metadata import ActorMetadata

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=ActorMetadata)})


class FaceBoundingBoxModalityContainer(containers.DeclarativeContainer):

    from .face_bounding_box import FaceBoundingBox

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=FaceBoundingBox)})


class KeypointsModalityContainer(containers.DeclarativeContainer):

    from .keypoints import v1, v2

    create = providers.FactoryAggregate(
        {1: modality_dataclass_factory(clazz=v1.SceneKeypoints), 2: modality_dataclass_factory(clazz=v2.SceneKeypoints)}
    )


class DatapointRequestModalityContainer(containers.DeclarativeContainer):

    from datagen.api.assets import DataRequest

    create = providers.FactoryAggregate({1: providers.Callable(modality_pydantic_factory(clazz=DataRequest))})


class IdentitiesModalitiesContainer(BaseModalitiesContainer):
    actor_metadata = providers.Callable(
        modality_factory, modality_container=providers.Container(ActorMetadataModalityContainer)
    )

    face_bounding_box = providers.Callable(
        modality_factory, modality_container=providers.Container(FaceBoundingBoxModalityContainer)
    )

    keypoints = providers.Callable(modality_factory, modality_container=providers.Container(KeypointsModalityContainer))

    datapoint_request = providers.Callable(
        modality_factory, modality_container=providers.Container(DatapointRequestModalityContainer)
    )



from typing import List

import marshmallow_dataclass
from marshmallow import pre_load

from datagen.modalities.textual.common.ndarray import NumpyArray


@marshmallow_dataclass.dataclass
class Orientation:
    look_at_vector: NumpyArray
    up_vector: NumpyArray


@marshmallow_dataclass.dataclass
class LightSource:
    name: str
    brightness: float
    beam_angle: float
    location: NumpyArray
    orientation: Orientation
    type: str
    falloff: float
    spectrum: str


@marshmallow_dataclass.dataclass
class LightsMetadata:
    lights: List[LightSource]

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs) -> dict:
        lights = []
        for light_source_name, light_source_metadata_dict in in_data["lights"].items():
            lights.append({"name": light_source_name, **light_source_metadata_dict})
        return {"lights": lights}



from dataclasses import field
from typing import TypeVar, Optional

import marshmallow
import marshmallow_dataclass
from marshmallow import ValidationError
from marshmallow.fields import Field

from datagen.modalities.textual.common.ndarray import NumpyArray

SubSegments = TypeVar("SubSegments")


@marshmallow_dataclass.dataclass
class Keypoints:
    name: str
    coords_2d: NumpyArray = field(repr=False)
    coords_3d: NumpyArray = field(repr=False)
    is_visible: NumpyArray = field(repr=False)


class SubSegmentsField(Field):
    def _deserialize(self, value, *args, **kwargs):
        sub_segments = []
        for name, data in value.items():
            try:
                sub_segments.append(Keypoints.Schema().load({"name": name, **data}))
            except ValidationError:
                sub_segments.append(NestedSegment(name, self._deserialize(data, *args, **kwargs)))
        return sub_segments


class KeypointsSchema(marshmallow.Schema):
    TYPE_MAPPING = {SubSegments: SubSegmentsField}


@marshmallow_dataclass.dataclass(base_schema=KeypointsSchema)
class NestedSegment:
    name: str
    sub_segments: Optional[SubSegments]

    def __getattr__(self, item):
        for sub_seg in self.sub_segments:
            if sub_seg.name == item:
                return sub_seg

    def __dir__(self):
        return [seg.name for seg in self.sub_segments]


@marshmallow_dataclass.dataclass(base_schema=KeypointsSchema)
class SceneKeypoints:
    scene: SubSegments

    def __getattr__(self, item):
        for sub_seg in self.scene:
            if sub_seg.name == item:
                return sub_seg

    def __dir__(self):
        return [seg.name for seg in self.scene]



from datagen.core.tasks.containers import TaskContainer
from datagen.core.tasks.task import ClientTask, Task, TaskChain, TaskCollection, TaskGroup
from datagen.core.tasks.task_definitions import (
    DownloadFileTask,
    ExtractFilesTask,
    FinalizeDataGenerationTask,
    GetDownloadURLsTask,
    InitDataGenerationTask,
    StatusTask,
    StopGenerationTask,
    UploadRequestTask,
)
from datagen.core.tasks.task_runner import TaskRunner



from .entity.base import DataPoint
from .entity.hic import DataPoint as HICDataPoint
from .entity.identities import DataPoint as IdentitiesDataPoint
from .repo import DatapointsRepository
from .container import DatapointsContainer



from typing import Any, Optional

import aiohttp
from pydantic import BaseModel

DEFAULT_CHUNK_SIZE = 1024 * 1024  # 1MB
DEFAULT_SESSION_TIMOUT_SEC = 300  # 5 min


class SessionsResponse(BaseModel):
    status_code: int
    payload: Any


class ClientSession:
    def __init__(
        self,
        base_url: Optional[str] = None,
        headers: Optional[dict] = None,
        timeout: Optional[float] = DEFAULT_SESSION_TIMOUT_SEC,
    ):
        self.base_url = base_url
        self.headers = headers
        self.timeout = aiohttp.ClientTimeout(total=timeout)
        self.session = None

    async def __aenter__(self):
        self.session = aiohttp.ClientSession(base_url=self.base_url, headers=self.headers, timeout=self.timeout)
        return self.session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.session.close()
        self.session = None



from dataclasses import dataclass, field
from pathlib import Path
from typing import Tuple

import cv2

from datagen.components.datapoint import DataPoint
from datagen.modalities.textual.base.environments import Environment


@dataclass
class Sequence:
    scene_path: Path
    camera_name: str
    datapoints: Tuple[DataPoint] = field(repr=False)
    environment: Environment = field(init=False)

    def __post_init__(self):
        self.environment = self._get_sequence_env()

    def _get_sequence_env(self) -> Environment:
        try:
            sequence_envs = set(map(lambda dp: dp.environment, self.datapoints))
            [environment] = list(sequence_envs)
            return environment
        except ValueError:
            # Too many values to unpack
            raise ValueError(f"More than one environment selected for sequence: {sequence_envs}")

    def __iter__(self):
        for datapoint in self.datapoints:
            yield datapoint

    def to_video(
            self, video_name: str = None, fps: int = 30, codec: str = "MJPG", height: int = None, width: int = None
    ) -> cv2.VideoWriter:
        """
        :param video_name Name of the video. Must also include format. If not specified,
        use the sequence's behaviour name, and .mov format
        :param fps:
        :param codec:
        :param height Height of the video. If not specified, use same as the datapoints'
        :param width Width of the video. If not specified, use same as the datapoints'
        """
        video = self._create_video(video_name, fps, codec, height, width)
        self._write_images(video)
        video.release()
        return video

    def _create_video(self, video_name: str, fps: int, codec: str, height: int, width: int) -> cv2.VideoWriter:
        if height is None or width is None:
            height, width = self._get_datapoints_shape()
        if video_name is None:
            video_name = f"{self.environment.behaviour}.mov"
        vid = cv2.VideoWriter(video_name, cv2.VideoWriter_fourcc(*codec), fps, (width, height))
        vid.release()
        return vid

    def _get_datapoints_shape(self) -> Tuple[int, int]:
        height, width, _ = self.datapoints[0].visible_spectrum.shape
        return height, width

    def _write_images(self, video: cv2.VideoWriter) -> None:
        for dp in self.datapoints:
            video.write(dp.visible_spectrum)



from dataclasses import dataclass
from typing import List, Optional, Union

from datagen.api.assets import Accessories, Background, Camera, Glasses, Human, HumanDatapoint, Light, Mask


@dataclass
class HumanDatapointBuilder:
    human: Human
    camera: Camera
    glasses: Optional[Glasses]
    mask: Optional[Mask]
    background: Optional[Background]
    lights: Optional[List[Light]]

    def get_basic_datapoint(self) -> HumanDatapoint:
        return HumanDatapoint(human=self.human.copy(deep=True), camera=self.camera.copy(deep=True))

    def get_accessories(self) -> Union[Accessories, None]:
        if self.glasses is None and self.mask is None:
            return None
        else:
            accessories = Accessories()
            if self.glasses is not None:
                accessories.glasses = self.glasses.copy(deep=True)
            if self.mask is not None:
                accessories.mask = self.mask.copy(deep=True)
            return accessories

    def get_background(self) -> Union[Background, None]:
        if self.background is not None:
            return self.background.copy(deep=True)
        else:
            return None

    def get_lights(self) -> Union[List[Light], None]:
        if self.lights is not None and len(self.lights) > 0:
            return [l_.copy(deep=True) for l_ in self.lights]
        else:
            return None



from typing import List

import marshmallow_dataclass
from marshmallow import pre_load

from datagen.modalities.textual.common.identity_label import IdentityLabel
from datagen.modalities.textual.common.sixdof import SixDOF


@marshmallow_dataclass.dataclass
class ActorMetadata:
    identity_id: str
    identity_label: IdentityLabel
    head_six_dof: SixDOF


@marshmallow_dataclass.dataclass
class Actors:
    identities: List[ActorMetadata]

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        identities = []
        for actor_identity_id, actor_metadata in in_data["identities"].items():
            actor_metadata = {
                "identity_id": actor_identity_id,
                "identity_label": {
                    "age": actor_metadata["age"],
                    "ethnicity": actor_metadata["ethnicity"],
                    "gender": actor_metadata["gender"],
                },
                "head_six_dof": actor_metadata["head_six_dof"],
            }
            identities.append(actor_metadata)
        return {"identities": identities}

    def __getitem__(self, key):
        return next(identity for identity in self.identities if identity.identity_id == key)

    def __iter__(self):
        for identity in self.identities:
            yield identity



from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from dependency_injector.wiring import inject, Provide

from datagen.components.datapoint import DataPoint
from datagen.components.datapoint import DatapointsRepository
from datagen.components.sequence import Sequence


@dataclass
class Camera:
    name: str
    scene_path: Path
    datapoints: List[DataPoint] = field(init=False, repr=False)

    def __post_init__(self):
        self.datapoints = self._init_datapoints()

    @inject
    def _init_datapoints(self, repo: DatapointsRepository = Provide["repo"]) -> List[DataPoint]:
        return repo.get_datapoints(scene_path=self.scene_path, camera_name=self.name)

    def get_sequence(self, **env_attributes) -> Sequence:
        """"
        :returns a sequence of datapoints ordered by frames, in context of a single environment (Time of Day etc.)
        """
        return Sequence(
            scene_path=self.scene_path,
            camera_name=self.name,
            datapoints=tuple(dp for dp in self.datapoints if dp.environment.matches(**env_attributes)),
        )

    def __iter__(self):
        for datapoint in self.datapoints:
            yield datapoint

    def __getitem__(self, key):
        return self.datapoints[key]

    def __len__(self):
        return len(self.datapoints)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['datagen',
 'datagen.api',
 'datagen.api.catalog',
 'datagen.api.catalog.hooks',
 'datagen.api.client',
 'datagen.api.datapoint',
 'datagen.api.requests',
 'datagen.api.requests.datapoint',
 'datagen.components',
 'datagen.components.datapoint',
 'datagen.components.datapoint.entity',
 'datagen.config',
 'datagen.core',
 'datagen.core.tasks',
 'datagen.dev',
 'datagen.imaging',
 'datagen.modalities',
 'datagen.modalities.textual',
 'datagen.modalities.textual.base',
 'datagen.modalities.textual.common',
 'datagen.modalities.textual.hic',
 'datagen.modalities.textual.identities',
 'datagen.modalities.textual.identities.keypoints',
 'datagen.utilities']

package_data = \
{'': ['*'], 'datagen.api.catalog': ['cache/*', 'cache/hic/*', 'cache/humans/*']}

install_requires = \
['aiofiles>=23.1.0,<24.0.0',
 'aiohttp>=3.8,<4.0',
 'appdirs>=1.4,<2.0',
 'asyncio-channel>=0.9.1,<0.10.0',
 'datagen-protocol-functional==0.8.0',
 'dependency-injector>4.6',
 'dynaconf>=2.2.3,<3.0.0',
 'importlib-metadata<5.0.0',
 'marshmallow-dataclass>=6',
 'marshmallow-enum>=1.5',
 'nest-asyncio>=1.5.6,<2.0.0',
 'opencv-python>=4.4,<5.0',
 'pandas>=1.1,<2.0',
 'pydantic>=1.9,<2.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=12.5,<13.0']

extras_require = \
{':python_full_version >= "3.6.7" and python_version < "3.7"': ['numpy>=1.15,<2.0'],
 ':python_version >= "3.7" and python_version < "3.8"': ['numpy>=1.19,<2.0'],
 ':python_version >= "3.8" and python_version < "3.9"': ['numpy>=1.20,<2.0'],
 ':python_version >= "3.9"': ['numpy>=1.20,<2.0'],
 'camera-utils:python_version < "3.9"': ['scipy>=1.5,<2.0', 'scipy>=1.5,<2.0'],
 'camera-utils:python_version >= "3.9" and python_version < "4.0"': ['scipy>=1.9,<2.0',
                                                                     'scipy>=1.9,<2.0']}

setup_kwargs = {
    'name': 'datagen-tech',
    'version': '1.8.1rc10',
    'description': 'Datagen SDK',
    'long_description': '# The Datagen SDK Python Package\n\n## About Datagen\n\nDatagen is powering the AI revolution by providing high-performance synthetic data, with a focus on data for human-centric computer vision applications. \n\nDatagen provides a powerful platform that allows you to generate high-quality and high variance, domain-specific, simulated synthetic data, granting you the ability to simulate dynamic humans and objects in their context. With Datagen, CV engineers have unparalleled flexibility to control visual outcomes across a broad spectrum of 3D environments.\n\n## About this package\n\nDatagen has developed a set of Python-based user tools for accessing and loading the rendered images and metadata that you generate on our platform. This package is designed to load our synthetic data into your Python environment, whether as part of your training set or as part of your testing set. \n\nThe abstraction layer we designed gives you access to all of the modalities we offer: facial landmarks, eye gaze vectors, depth and normal maps, and more. \n\n## Installation\n\nBefore installing this package, we recommend that you create a new Conda environment using the following command:\n\n> conda create --name datagen-env\n\nNext, activate the environment using this command:\n\n> conda activate datagen-env\n\nFinally, use this command to install the Datagen package itself:\n\n> pip install datagen-tech\n\n# Using This Package\n\nThis package contains Python objects and functions that are designed to process datasets generated with the Datagen Platform. If you are a Datagen customer, you have already been given access to sample datasets and jupyter notebook tutorials. If you are not a Datagen customer, contact our support team at support@datagen.tech.\n\n## Importing the SDK\n\nTo import the sdk, use the command `import datagen` or, if you prefer, `import datagen as dg`. For the rest of this tutorial we will use the latter.\n\n## Datasets\n\nA "dataset" is any collection of datapoints that were generated on the Datgen platform. Each datapoint is a single visual spectrum image accompanied by visual and textual metadata.\n\n### Loading a dataset\n\nThe `dg.load()` command stores one or more datasets into a `Dataset` object. It can accept any number of target folders at once, merging them into one large dataset so you can conduct operations on all of them together. \n\nThis command is fully backwards-compatible. If you use the platform\'s newest features in your dataset, you can still load that dataset together with older datasets that predate those features.\n\nWhen using this command, you should target the top-level folder of each dataset (the folder that contains one or more subfolders named "scene").  For example:\n\n```\ndataset = dg.load("/path/to/datagen/datasets/eye-gaze-forward", "/path/to/datagen/datasets/expression-smiles")\n```\n\nThe above line of sample code loads a dataset found in a folder named "eye-gaze-forward" as well as a dataset found in a folder named "expression-smiles", merging their datapoints into a single dataset.\n\n### Working with your dataset\n\nYou can iterate over your dataset in three different ways: by scene, by camera, or by datapoint.\n\n### Scenes\n\n#### What is a scene?\n\nA "scene" refers to a 3D environment that contains one or more 3D objects. For example, each scene that is created in our Faces generator consists of a single human face, wearing a specific expression, and located in a static position.\n\nWhen generating a dataset on our Faces platform, you have the option of rendering each scene in multiple ways: through more than one camera, each with different settings; and under more than one lighting scenario, each with different backgrounds. Therefore, depending on the settings you used when you generated this dataset, each scene can be depicted in anywhere between one and over 30 rendered images.\n\n#### The `Scene` object\n\nEach dataset contains an iterable set of `Scene` objects. To retrieve this set, use `dataset.scenes`. To refer to an individual scene in a dataset, use `dataset.scenes[0]`.\n\nThe `Scene` object contains a set of rendered images that all depict the same subject, but from different angles and under different lighting scenarios. Each of these images, along with its metadata, is called a "datapoint". You can access a scene\'s datapoints by subscription:\n\n```\nscene = dataset.scenes[0]\ndatapoint = scene[0]\n```\n\nOr by iterating over the scene:\n\n```\nfor datapoint in scene:\n    ...\n```\n\n**Note:** In older datasets generated by our platform, scenes were previously named "environments". The package supports both formats.\n\n### Cameras\n\n#### What is a camera?\n\nA "camera" refers to a set of rendered images that were taken of a specific scene, from a specific angle, under specific camera settings - but each one shows the scene under a different lighting scenario. When you generated this dataset, you selected one or more lighting scenarios; that selection determines the number of rendered images per camera.\n\n#### The `Camera` object\n\nEach scene contains an iterable list of `Camera` objects. To retrieve this set, use scene`scene.cameras`. To refer to an individual camera in a scene, use `scene.cameras[0]`.\n\nThe `Camera` object contains a set of rendered images that all depict the same subject from the same angle, but under different lighting scenarios. Each of these images, along with its metadata, is called a "datapoint". You can access a camera\'s datapoints by subscription:\n\n```\nscene = dataset.scenes[0]\ncamera = scene.cameras[0]\ndatapoint = camera[0]\n```\n\nOr by iterating over the camera:\n\n```\nfor datapoint in camera:\n    ...\n```\n\n### Datapoints\n\n#### What is a datapoint?\n\nA "datapoint" refers to a single rendered image of a synthetic human subject, along with all of its metadata. That metadata includes all of Datagen\'s modalities: normal and depth maps, facial landmark coordinates, camera and actor metadata, and more.\n\n#### The `Datapoint` object\n\nA `Datapoint` object consists of a visual spectrum image annotated by additional visual and textual data files:\n\n* The visual spectrum image depicts the scene from the point of view of a specific camera under a specific lighting scenario.\n\n* The annotations contain metadata that is specific to that image.\n\nFor example, each `Datapoint` object contains 2D coordinates for facial landmarks, identifying where those landmark can be found in the visual spectrum image.\n\nYou can use tab autocompletion to view the full list of available objects in the `Datapoint` object. For example, you can use the following command to access the visual spectrum image of the subject:\n\n```\nimshow(datapoint.visible_spectrum)\n```\n\nAnd this one to access a normal map of the subject:\n\n```\nimshow(datapoint.normals_map)\n```\n\nTextual annotations are organized in a hierarchy. `datapoint.actor_metadata.face_expression` gives you the description of the subject\'s facial expression, while `datapoint.semantic_segmentation_metadata.nose` gives you the RGB value that represents the nose in the accompanying semantic segmentation map.\n',
    'author': 'Datagen Technologies Ltd.',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/datagen1/core/sdk',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field

from datagen.api.assets import HumanDatapoint
from datagen.api.requests.datapoint.builder import HumanDatapointBuilder


@dataclass
class DataRequestDirector:
    builder: HumanDatapointBuilder = field(init=False)

    def build_datapoint(self) -> HumanDatapoint:
        datapoint = self.builder.get_basic_datapoint()
        datapoint.accessories = self.builder.get_accessories()
        datapoint.background = self.builder.get_background()
        datapoint.lights = self.builder.get_lights()
        return datapoint.copy(deep=True)



import sys

from datagen.dev import FunctionalModule
from datagen.api.catalog.containers import AssetsCatalogContainer

catalog = AssetsCatalogContainer().catalog()

sys.modules[__name__] = FunctionalModule(functionality=catalog)



import sys

from datagen.api import catalog  # noqa: F401
from datagen.api.containers import ApiContainer
from datagen.api.impl import DatagenAPI
from datagen.dev import FunctionalModule

try:
    from datagen.utilities import camera as camera_utils  # noqa: F401
except ModuleNotFoundError:
    # Did not install the camera-utils optional dependency
    pass

api = ApiContainer.api()

sys.modules[__name__] = FunctionalModule(functionality=api)
